package si.JDEXi;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

/**
 * ComplexRules class. Represents complex rule algorithm from DEXi. DEXi uses complex rules in order to 
 * present its utility functions in a more compact and possibly more comprehensible way than with elementary
 * rules which have the same function value. Complex rules are obtained by joining several elementary
 * rules which have the same function value. In other words, a complex rule represents one or more
 * elementary rules. In tables, complex rules are characterized by the use of intervals.
 * This algorithm includes improvements compared to original algorithm implemented in DEXi software.
 * 
 * @author Adem Kikaj, Marko Bohanec
 * @version 4.0
 * @since 2018-02-15
 */

public class ComplexRules 
{
	
	/**
	* Selected attribute for creating complex rules.
	*/
	private Attribute attribute;
	/**
	* Integer of rule length.
	*/
	private int ruleLength;
	/**
	* List of elementary rules.
	*/
	private ArrayList<CRule> elementaryRules;
	/**
	* List of temporary elementary rules.
	*/
	private ArrayList<CRule> tmpElementaryRules;
	/**
	* List of elementary rules with same target.
	*/
	private ArrayList<CRule> targetedRules;
	/**
	* List of highest value allowed of all arguments.
	*/
	private ArrayList<String> arguments;
	/**
	* List of complex rules.
	*/
	private ArrayList<String> complexRules;
	/**
	* List of complex rules encoded for display.
	*/
	private ArrayList<String> dispCR;
	/**
	* List of distinct complex rules.
	*/
	private ArrayList<CRule> cleanedComplexRules;
	/**
	* List of rules that have been covered by algorithm.
	*/
	private ArrayList<String> alreadyTried;
	/**
	* List of complex rules.
	*/
	private ArrayList<String> stringCCR; 
	/**
	* Separated line between complex rules with different targets.
	*/
	private String seperateLine;
	/**
	* Rule target.
	*/
	private char target;
	/**
	* StringBuffer used to increment and decrement new low and high boundaries.
	*/
	private StringBuffer nh,nl;
	/**
	* Integer of complex rules counter.
	*/
	private int ruleCount = 1;
	/**
	* Boolean used to check inconsistent rules.
	*/
	private Boolean inconsistent;	
	/**
	* Boolean used to check if the rules should be printed with colors in terminal.
	*/
	private Boolean colorOutput;
	
	
	/**
	 * Constructor with selected attribute and colorOutput parameters.
	 * 
	 * @param attribute 	    
	 *            Attribute object parameter    
	 * @param colorOutput 
	 *            Boolean object parameter            
	 */
	public ComplexRules(Attribute attribute, Boolean colorOutput)
	{
		this.attribute=attribute;
		this.ruleLength = 0;
		this.elementaryRules = new ArrayList<CRule>();
		this.tmpElementaryRules = new ArrayList<CRule>();
		this.targetedRules = new ArrayList<CRule>();
		this.arguments = new ArrayList<String>();
		this.complexRules = new ArrayList<String>();	
		this.dispCR = new ArrayList<String>();
		this.cleanedComplexRules  = new ArrayList<CRule>();
		this.stringCCR = new ArrayList<String>();
		this.alreadyTried = new ArrayList<String>();
		this.inconsistent = false;		
		this.seperateLine = "";
		this.colorOutput = colorOutput;
	}
	
	/**
	 * Generates and print complex rules for each target.
	 * 
	 */
	public void createComplexRule()
	{
		try {
			this.ruleLength = attribute.indexToArgs(0).length;
			addArgumentScale();
			addElementaryRules();	
			CreateSeperateLine();
			for(int i=0; i<this.attribute.getScaleSize(); i++)
			{
				this.target = digitToChar(i);
				addTargetedRules(this.target);
				if(this.inconsistent) 
				{
					copyArray();					
					addElementaryInconsistenRules();		
					sortElementaryRules();
					sortTargetedRules();
				}
				generateComplexRules();				
				this.targetedRules.clear();
				findComplexRules();
				displayGeneratedRules();
				if(this.stringCCR.size()>0)
					System.out.println(seperateLine);
				this.cleanedComplexRules.clear();
				this.stringCCR.clear();
				if(this.inconsistent)
				{
					this.elementaryRules.clear();
					for(int j=0; j<this.tmpElementaryRules.size(); j++)
						this.elementaryRules.add(new CRule(this.tmpElementaryRules.get(j).getRule(), false));
					this.tmpElementaryRules.clear();
				}
			}
			
		} catch (Exception e) {
			System.out.println(e);
		}		
	}
	
	/**
	 * Creates separate line based on the number of arguments and length of the 
	 * named attributes together with target.
	 * 
	 */
	private void CreateSeperateLine()
	{
		for(int i=0; i<=this.attribute.getAttributeCount()-1; i++)
		{
			seperateLine += "---------------";
		}
		int l = this.attribute.getScaleValue(0).length();
		for(int i=1; i<this.attribute.getScaleSize(); i++)
		{
			String s = this.attribute.getScaleValue(i);
			if(l < s.length())
				l = s.length();
		}
		for(int i=0; i<l+7; i++)
		{
			seperateLine += "-";
		}
	}
	
	/**
	 * Copy elementary rules to new array.
	 * 
	 */
	private void copyArray()
	{
		for(int i=0; i<this.elementaryRules.size(); i++)
		{
			this.tmpElementaryRules.add(new CRule(this.elementaryRules.get(i).getRule(), false));
		}
		
	}
	
	/**
	 * Sorts targeted rules which are used for complex algorithm to start 
	 * from the lowest bound.
	 */
	private void sortTargetedRules() 
	{
		ArrayList<String> tmp = new ArrayList<String>();
		for(int i=0; i<this.targetedRules.size(); i++)
		{
			tmp.add(this.targetedRules.get(i).getRule());
		}
		Collections.sort(tmp);
		this.targetedRules.clear();
		for(int i=0; i<tmp.size(); i++)
		{
			this.targetedRules.add(new CRule(tmp.get(i), false));
		}
	}
	
	/**
	 * Sorts elementary rules after inconsistent rules are added. 
	 */
	private void sortElementaryRules()
	{
		ArrayList<String> tmp = new ArrayList<String>();
		for(int i=0; i<this.elementaryRules.size(); i++)
		{
			tmp.add(this.elementaryRules.get(i).getRule());
		}
		Collections.sort(tmp);
		this.elementaryRules.clear();
		for(int i=0; i<tmp.size(); i++)
		{
			this.elementaryRules.add(new CRule(tmp.get(i), false));
		}
	}
	
	/**
	 * Creates array of complex rules that are only covered correctly and 
	 * prepares this array for output
	 */
	private void findComplexRules()
	{
		String r,l,h;
		for(int i=0; i<this.stringCCR.size(); i++)
		{
			this.cleanedComplexRules.add(new CRule(this.stringCCR.get(i), false));
		}		
		for(int i=0; i<this.cleanedComplexRules.size(); i++)
		{
			if(!this.cleanedComplexRules.get(i).isFound()) {
				r = this.cleanedComplexRules.get(i).getRule();
				l = r.substring(0, this.ruleLength);
				h = r.substring(this.ruleLength,this.ruleLength*2);
				if(this.complexRules.indexOf(l)<0)
					this.complexRules.add(l);	
				if(this.complexRules.indexOf(h)<0)
					this.complexRules.add(h);	
				this.cleanedComplexRules.get(i).setFound(true);
				findCoveringComplexRules(l,h);
				clear();
			}
		}	
		
	}
	
	/**
	 * For each generated complex rule which includes
	 * low and high bound, an array is created only 
	 * with the elementary rules that each complex rule covers
	 * which is used to manipulate this covered elementary rules 
	 * for finding intervals and printing them.
	 * 
	 * @param l
	 * @param h
	 */
	private void findCoveringComplexRules(String l, String h)
	{
		String nr,nl,nh;
		boolean covering = false;
		for(int i=0; i<this.cleanedComplexRules.size(); i++)
		{
			if(!this.cleanedComplexRules.get(i).isFound())
			{
				nr = this.cleanedComplexRules.get(i).getRule();
				nl = nr.substring(0, this.ruleLength);
				nh = nr.substring(this.ruleLength,this.ruleLength*2);
				for(int j=0; j<l.length(); j++)
				{
					if(charToDigit(l.charAt(j)) <= charToDigit(nl.charAt(j)) && charToDigit(h.charAt(j)) >= charToDigit(nh.charAt(j)))
						covering = true;
					else
					{
						covering = false;
						break;
					}
				}
				if(covering) {
					if(this.complexRules.indexOf(nl)<0)
						this.complexRules.add(nl);	
					if(this.complexRules.indexOf(nh)<0)
						this.complexRules.add(nh);	
					this.cleanedComplexRules.get(i).setFound(true);
				}
			}
		}
	}
	
	/**
	 * For each attribute array arguments is assigned as 
	 * highest scale value of that attribute. 
	 */
	private void addArgumentScale()
	{
		for (int i = 0; i < attribute.getAttributeCount(); i++) 
		{
			Attribute a = attribute.getAttribute(i);
			this.arguments.add((a.getScaleSize()-1) + "");
		}
	}
	
	/**
	 * Creates elementary rules of attribute from 
	 * utility function as complex rules which are used
	 * to generate complex rules.
	 */
	private void addElementaryRules()
	{
		String tmp="";
		for (int i = 0; i < attribute.getRuleCount(); i++) 
		{
            Rule rule = attribute.getRule(i);
            int[] args = attribute.indexToArgs(i);
            if(rule.getLow().equals(rule.getHigh()))
            {
            	for(int j=0; j<args.length; j++)
            	{
            		tmp += args[j];
            	}
                elementaryRules.add(new CRule(tmp+rule.getLow(), false)); 
                tmp = "";
            }
            else            
            	this.inconsistent = true;
            
        }
	}

	/**
	 * Creates rules as elementary when they are targeted with intervals.
	 * 
	 */
	private void addElementaryInconsistenRules()
	{
		String tmp="";
		for (int i = 0; i < attribute.getRuleCount(); i++) 
		{
            Rule rule = attribute.getRule(i);
            int[] args = attribute.indexToArgs(i);
            if(!rule.getLow().equals(rule.getHigh()))
            {
            	for(int j=0; j<=rule.getHigh();)
            	{
                    tmp = "";
            		for(int k=0; k<args.length; k++)
                	{
                		tmp += args[k];
                	}
            		if(rule.getLow()==charToDigit(this.target))
            		{
            			String theRule = tmp;
            			this.elementaryRules.add(new CRule(theRule+rule.getLow(), false));
            			this.targetedRules.add(new CRule(theRule+rule.getLow(), false));
            			break;
            		}
            		else if(rule.getHigh()==charToDigit(this.target))
            		{
            			String theRule = tmp;
            			this.targetedRules.add(new CRule(theRule+rule.getHigh(), false));   
            			this.elementaryRules.add(new CRule(theRule+rule.getHigh(), false));
            			break;
            		}
            		else if(charToDigit(this.target) != rule.getLow() && charToDigit(this.target) > rule.getLow() && charToDigit(this.target) < rule.getHigh())
            		{
            			String theRule = tmp;
            			this.targetedRules.add(new CRule(theRule+charToDigit(this.target), false));   
            			this.elementaryRules.add(new CRule(theRule+charToDigit(this.target), false));
            			break;
            		}
            		else
            		{
            			String theRule = tmp;
            			this.elementaryRules.add(new CRule(theRule+rule.getLow(), false));
            			break;
            		}
            			
            	}
            }
		}
		
	}
	
	/**
	 * For specific target it creates array with targeted
	 * elementary rules in order join these rules.
	 * 
	 */
	private void addTargetedRules(char target)
	{
		String tmpRule="";
		int tmpRuleLength = 0;
		for(int i=0; i<this.elementaryRules.size(); i++)
		{
			tmpRule = this.elementaryRules.get(i).getRule();
			tmpRuleLength = tmpRule.length()-1;
			if(tmpRule.charAt(tmpRuleLength)==target) { 
				this.targetedRules.add(new CRule(tmpRule, false));
			}
		}
	}
	
	/**
	 * For each elementary rule with same target we try to 
	 * expand them as much as we can and if any of these
	 * targeted rules are not covered before we do it again for 
	 * non covered rules.
	 */
	private void generateComplexRules()
	{
		String targetedRule;
		for(int i=0; i<this.targetedRules.size(); i++)
		{
			if(!this.targetedRules.get(i).isFound()) {
				targetedRule = this.targetedRules.get(i).getRule();
				this.targetedRules.get(i).setFound(true);
				Cover(targetedRule,targetedRule);
				alreadyTried.clear();
			}
		}
	}	
		
	/**
	 * Starts with a specifically targeted rule where l as low and h as high 
	 * are the same rule in the beginning and tries to expand them as much as it can.
	 * 
	 * @param l String object parameter
	 * @param h String object parameter	 
	 */
	private boolean Cover(String l,String h)
	{		
		boolean complexRule;
		boolean NextCover = false;
		if(alreadyTried.indexOf(l+h) == -1) {
			alreadyTried.add(l+h);
			complexRule = CoveringClass(l,h);			
			if(!complexRule){ 
				return false;
			}
			if(!NextCover) 
			{
				found(l);	
				found(h);	
			}
			for(int i=0; i<this.ruleLength; i++)
			{
				if(charToDigit(h.charAt(i))-charToDigit(arguments.get(i).charAt(0))<0) {
					this.nh = new StringBuffer(h);
					this.nh.replace(i, i+1, charToDigitInc(nh.charAt(i)));	
						if (Cover(l,nh.toString()))
							NextCover = false;					
				}				
			}		
			for(int i=0; i<this.ruleLength; i++)
			{
				if(charToDigit(l.charAt(i))>0)
				{
					this.nl = new StringBuffer(l);
					this.nl.replace(i, i+1, charToDigitDec(nl.charAt(i)));
						if (Cover(nl.toString(),h))
							NextCover = true;						
				}
			}
			if(!NextCover) {
				String theRule = l.substring(0,this.ruleLength)+h.substring(0,this.ruleLength);
				if(this.stringCCR.indexOf(theRule)<0) {
					//System.out.println(theRule);
					this.stringCCR.add(theRule);
				}
			}
		}
		return false;
	}
	
	/**
	 * Check into elementary rules if the bound l - low and h - high
	 * Covers correctly specific target. 
	 * 
	 * @param l String object parameter
	 * @param h String object parameter	
	 * @return True if bounds covers correctly specific target. False if bounds covers other target.
	 */
	private boolean CoveringClass(String l, String h)
	{
		String s="";
		String s1 ="";
		boolean c=true;
		boolean result = false;
		for(int i=0; i<this.elementaryRules.size(); i++)
		{
			s1 = this.elementaryRules.get(i).getRule();
			s = this.elementaryRules.get(i).getRule().substring(0, this.ruleLength);
			c = true;
			for(int k=0; k<s.length(); k++)
			{
				c = true;
				if(!(s.charAt(k)>=l.charAt(k) && s.charAt(k)<=h.charAt(k)))
				{
					c = false;
					break;
				}				
			}
			if(c)		
			{
				result = (s1.charAt(s1.length()-1)==this.target);	
				if(!result)
					return result;
			}
		}
		return result;
		
	}
	
	/**
	 * Change the found value of specific rule S.
	 * @param s
	 */
	private void found(String s)
	{
		for(int i=0; i<this.targetedRules.size(); i++)
		{
			if(!this.targetedRules.get(i).isFound() && this.targetedRules.get(i).getRule().equals(s)) 
				this.targetedRules.get(i).setFound(true);						
		}
	}
	
	/**
	 * Founded complex rules encoded with numbers clear method translate them 
	 * into String and prepares for output.
	 */
	private void clear()
	{
		ArrayList<String> tmpComplexRules = new ArrayList<String>();
		String tmpRule = "";
		String complexRule = "";
		String formatedComplexRule = "";
		String[] tmpResult;
		boolean addInto = true;
		List<String> distinctList = this.complexRules.stream().distinct().collect(Collectors.toList());
		Collections.sort(distinctList);
		if(distinctList.size() > 0) 
		{
			for(int i = 0; i < distinctList.get(0).length(); i++)
	    	{
	      	    for(int j = 0; j < distinctList.size(); j++)
	            {
	      	    	tmpRule+=distinctList.get(j).charAt(i);
	            }
	            tmpComplexRules.add(tmpRule);
	            tmpRule = "";
	        }	
			for(int i=0; i<tmpComplexRules.size(); i++)
			{
				tmpRule = tmpComplexRules.get(i);
				tmpRule = uniqueCharacters(tmpRule);
				if(tmpRule.length() == this.attribute.getAttribute(i).getScaleSize())
				{
					complexRule = "*";
				}
				else if(tmpRule.length() == 1)
				{
					if(Integer.parseInt(tmpRule.charAt(0)+"") == 0 && this.colorOutput) {
						complexRule = this.attribute.getAttribute(i).getScaleValue(Integer.parseInt(tmpRule.charAt(0)+""));	
						complexRule = String.format("%1$-24s", "\033[31m"+complexRule.toString()+ "\033[0m");
					}
					else if(Integer.parseInt(tmpRule.charAt(0)+"") == this.attribute.getAttribute(i).getScaleSize()-1 && this.colorOutput) {
						complexRule = this.attribute.getAttribute(i).getScaleValue(Integer.parseInt(tmpRule.charAt(0)+""));	
						complexRule = String.format("%1$-24s", "\033[32m"+complexRule.toString()+ "\033[0m");						
					}
					else
						complexRule = this.attribute.getAttribute(i).getScaleValue(Integer.parseInt(tmpRule.charAt(0)+""));	
				}
				else
				{
					tmpResult = findMinAndMax(tmpRule);
					int min = Integer.parseInt(tmpResult[0]);
					int max = Integer.parseInt(tmpResult[1]);
					if(min == 0 && charToDigit(tmpRule.charAt(tmpRule.length()-1)) < tmpRule.length()) // 012345 -> 012 => "<=2" 
					{
						complexRule = "<="+this.attribute.getAttribute(i).getScaleValue(max);
					}
					else if(max == (this.attribute.getAttribute(i).getScaleSize()-1) && (tmpRule.length() < this.attribute.getAttribute(i).getScaleSize()) && charToDigit(tmpRule.charAt(0)) != 0)
					{
						complexRule = ">="+this.attribute.getAttribute(i).getScaleValue(Integer.parseInt(tmpResult[0]));
					}
					else if(min==0 && max==this.attribute.getAttribute(i).getScaleCount()-1)
					{
						complexRule = "*";
					}
					else if(min==0 && charToDigit(tmpRule.charAt(tmpRule.length()-1)) == (tmpRule.length()))
					{
						complexRule = "<="+this.attribute.getAttribute(i).getScaleValue(max);
					}
					else 
					{
						complexRule = this.attribute.getAttribute(i).getScaleValue(min) + ":"+ this.attribute.getAttribute(i).getScaleValue(max);
					}
				}
				formatedComplexRule += String.format("%1$-15s", complexRule);
				tmpRule = "";		
			}
			if(Integer.parseInt(this.target+"") == 0 && this.colorOutput) {
				String t = "\033[31m"+this.attribute.getScaleValue(Integer.parseInt(this.target +""))+ "\033[0m";
				formatedComplexRule += "|"+t+"|";
			}
			else if(Integer.parseInt(this.target+"") == this.attribute.getScaleSize()-1 && this.colorOutput) {
				String t = "\033[32m"+this.attribute.getScaleValue(Integer.parseInt(this.target +""))+ "\033[0m";
				formatedComplexRule += "|"+t+"|";
			}
			else
				formatedComplexRule += "|"+this.attribute.getScaleValue(Integer.parseInt(this.target +""))+"|";
			this.complexRules.clear();	
			if(addInto) {
				this.dispCR.add(formatedComplexRule);
			}
		}	
	}
	
	/**
	 * For specific complex rules it finds minimum and maximum
	 * value used to translate into interval for complex rules.
	 * Example: String s = 123 => return 1 as min and 3 as max, if for this
	 * argument 4 is the Scale Size it outputs this rule as >=1
	 * @param s
	 * @return Min and Max value as array
	 */
	private String[] findMinAndMax(String s)
	{
		int value = charToDigit(s.charAt(0));
		int maxValue = value;
		int minValue = value;		
		for(int i=1;i < s.length(); i++)
		{ 
			value = charToDigit(s.charAt(i));
			if(value > maxValue)
				maxValue = value;		      
			if(value < minValue)
		        minValue = value;		      
		} 
		return new String[]{minValue+"",maxValue+""};
	}
		
	/**
	 * All covered values for a specific Attribute are as input parameter.
	 * It returns unique values of this String.
	 * @param r
	 * @return String with unique characters
	 */
	private String uniqueCharacters(String r){
	    String temp = "";
	    for (int i = 0; i < r.length(); i++){
	        char current = r.charAt(i);
	        if (temp.indexOf(current) < 0){
	            temp += current;
	        }
	    }
	    if(temp.length() == 0)
	    	return r.charAt(0)+"";
	    else
	    	return temp;
	}

	/**
	 * Converts Number to Character.
	 * @param i
	 * @return
	 */
	private char digitToChar(int i)
	{
		return (char)(i+48);
	}
	
	/**
	 * Converts Character to Number
	 * @param c
	 * @return
	 */
	private int charToDigit(char c)
	{
		return (int)(c-48);
	}

	/**
	 * Converts Character to Number and it increments by one,
	 * then it returns this value as String.
	 * @param c
	 * @return
	 */
	private String charToDigitInc(char c)
	{
		int i = (int)(c-48);
		i++;
		return i+"";
	}
	
	/**
	 * Converts Character to Number and it decrement by one,
	 * then it returns this value as String.
	 * @param c
	 * @return
	 */
	private String charToDigitDec(char c)
	{
		int i = (int)(c-48);
		i--;
		return i+"";
	}
	
	/**
	 * Prints list of complex rules.
	 */
	private void displayGeneratedRules()
	{
		Collections.sort(this.dispCR);
		for(int i=0; i<this.dispCR.size(); i++)
		{			
			System.out.println(String.format("%1$-5s", ruleCount+") ")+this.dispCR.get(i));
			ruleCount++;
		}
		dispCR.clear();
	}
		
	}
