package si.JDEXi;

import java.util.ArrayList;
/**
 * NewComplexRules class. Represents complex rules in order to present its utility functions in 
 * a more compact and possibly more comprehensible way than with elementary rules which have 
 * the same function value. Complex rules are obtained by joining several elementary
 * rules which have the same function value. In other words, a complex rule represents one or more
 * elementary rules. In tables, complex rules are characterized by the use of intervals. Difference between
 * Complex Rules Algorithm is in the generated complex rules, and the robustness of this algorithm to 
 * find the complex rules from elementary ones. 
 * 
 * @author Adem Kikaj, Marko Bohanec
 * @version 4.0
 * @since 2018-02-15
 *
 */
public class NewComplexRules 
{
	/**
	* Selected attribute for creating complex rules.
	*/
	private Attribute attribute;
	/**
	 * List of Elementary Rules.
	 */
	private ArrayList<CRule> elementaryRules;
	/**
	 * List of targets for specific attribute.
	 */
	private ArrayList<String> targets;
	/**
	 * List of elementary rules with same target.
	 */
	private ArrayList<ArrayList<CRule>> groupedElementaryRules;
	/**
	 * List of lowest value all arguments for list of 
	 * grouped elementary rules.
	 */
	private ArrayList<String> lowestArguments; 
	/**
	 * Separated line between complex rules with different targets.
	 */
	private String seperateLine = ""; 
	/**
	 * Integer of complex rules counter.
	 */
	private int complexRulesCounter = 1; 
	
	/**
	 * Constructor with selected attribute parameter.
	 * @param attribute 
	 * 				Attribute object parameter
	 */
	public NewComplexRules(Attribute attribute)
	{
		//Initialize all variables 
		this.attribute = attribute;
		this.elementaryRules = new ArrayList<CRule>(); 
		this.targets = new ArrayList<String>();
		this.groupedElementaryRules = new ArrayList<ArrayList<CRule>>();
		this.lowestArguments = new ArrayList<String>();		
	}
	
	/**
	 * Generates and print complex rules for each target.
	 */
	public void createNewComplexRules()
	{
		AddElementaryRules();
		AddTargets();
		CreateSeperateLine();
		GroupByTarget();
		FindLowestArguments();
		FindComplexRules();
	}
	
	/**
	 * Creates separate line based on the number of arguments and length of the 
	 * named attributes together with target.
	 */
	private void CreateSeperateLine()
	{
		for(int i=0; i<=this.attribute.getAttributeCount()-1; i++)
		{
			seperateLine += "---------------";
		}
		int l = this.attribute.getScaleValue(0).length();
		for(int i=1; i<this.attribute.getScaleSize(); i++)
		{
			String s = this.attribute.getScaleValue(i);
			if(l < s.length())
				l = s.length();
		}
		for(int i=0; i<l+6; i++)
		{
			seperateLine += "-";
		}
	}
	
	/**
	 * Creates elementary rules of attribute from 
	 * utility function as complex rules which are used
	 * to generate complex rules.
	 */
	private void AddElementaryRules()
	{
		String tmp="";
		for (int i = 0; i < this.attribute.getRuleCount(); i++) 
		{
            Rule rule = attribute.getRule(i);
            int[] args = attribute.indexToArgs(i);
            for(int j=0; j<args.length; j++)        	
        		tmp += args[j];        	
            if(rule.getLow().equals(rule.getHigh()))                       	
                elementaryRules.add(new CRule(tmp+rule.getLow(), false));             
            else
            {
            	for(int j = rule.getLow(); j<=rule.getHigh(); j++)            	
            		elementaryRules.add(new CRule(tmp+j, false));             	
            }
            tmp = "";            
        }
	}

	/**
	 * Find all targets for this attribute.
	 */
	private void AddTargets()
	{
		for(int i=0; i<this.attribute.getScaleSize(); i++)
			this.targets.add(i + "");
	}

	/**
	 * Groups elementary rules by target and creates a 
	 * matrix.
	 */
	private void GroupByTarget()
	{
		for(int i=0; i<this.targets.size(); i++)
		{
			ArrayList<CRule> grouped = new ArrayList<CRule>();
			for(int j=0; j<this.elementaryRules.size(); j++)
			{
				String rule = this.elementaryRules.get(j).getRule();
				if(rule.substring(rule.length()-1, rule.length()).equals(i+""))
					grouped.add(new CRule(rule, false));
			}
			this.groupedElementaryRules.add(grouped);
		}
	}

	/**
	 * Creates an array with lowest arguments allowed for all 
	 * attributes from specific group of elementary rules.
	 */
	private void FindLowestArguments()
	{
		//For each grouped elementary rules, find lowest allowed argument.
		for(int i=0; i<groupedElementaryRules.size(); i++)
		{
			StringBuilder lowestArgument = new StringBuilder();
			String rule = groupedElementaryRules.get(i).get(0).getRule(); // assume that first rule is the lowest one
			rule = rule.substring(0, rule.length()-1); // get only rule without target	

			for(int j=0; j<rule.length(); j++)			
				lowestArgument.append(rule.charAt(j));			
			
			for(int j=1; j<groupedElementaryRules.get(i).size(); j++) //j==1 skip the first one 
			{
				rule = groupedElementaryRules.get(i).get(j).getRule();
				rule = rule.substring(0, rule.length()-1); 	
				for(int k=0; k<rule.length(); k++)
				{
					if(Integer.parseInt(lowestArgument.charAt(k)+"") > Integer.parseInt(rule.charAt(k)+""))
						lowestArgument.setCharAt(k, rule.charAt(k)); 
				}
			}
			this.lowestArguments.add(lowestArgument.toString());
		}
	}

	/**
	 * Generates complex rules.
	 */
	private void FindComplexRules()
	{
		String rule,high,low,target = "";
		CRule cRule;
		int nr,change;
		for(int i=0; i<this.groupedElementaryRules.size(); i++)
		{
			this.elementaryRules = this.groupedElementaryRules.get(i);
			for(int j=this.elementaryRules.size()-1; j>=0; j--)			
			{				
				cRule = this.elementaryRules.get(j);
				rule = cRule.getRule();
				target = rule.substring(rule.length()-1, rule.length());
				if(!cRule.isFound())
				{
					high = rule.substring(0, rule.length()-1);
					low = this.lowestArguments.get(i);
					ArrayList<Integer> combinations = FindCombinations(low,high);
					nr = NumberOfRules(combinations);
					change = combinations.get(combinations.size()-1);
					combinations = FixCombinations(high);
					while(Magic(low,change,nr,combinations,0)=="0")
					{
						low = Magic(low,change,nr,combinations,1);
						combinations = FindCombinations(low,high);
						nr = NumberOfRules(combinations);
						change = combinations.get(combinations.size()-1);
						combinations = FixCombinations(high);
					}
					ArrayList<String> complexRules = CountAndFind(low,change,nr,combinations);
					PrintComplexRules(complexRules,target,low.length());					
				}
			}
			System.out.println(seperateLine);
		}
	}
	
	/**
	 * Generates an array with number of combinations that can be
	 * between two rules respectively between low and high rule.
	 * @param low
	 * @param high
	 * @return List of Integers.
	 */
	private ArrayList<Integer> FindCombinations(String low, String high)
	{
		ArrayList<Integer> combinations = new ArrayList<Integer>();
		for(int i=0; i<low.length(); i++)
		{
			combinations.add((Integer.parseInt(high.charAt(i)+"")-Integer.parseInt(low.charAt(i)+""))+1);
		}
		return combinations;
	}

	/**
	 * Finds number of rules between two bounds.
	 * @param combinations
	 * @return Integer: number of rules between two bounds.
	 */
	private int NumberOfRules(ArrayList<Integer> combinations)
	{
		int nr = 1;
		for(int i=0; i<combinations.size(); i++)
		{
			nr *= combinations.get(i);
		}
		return nr;
	}

	/**
	 * Creates an array of highest allowed arguments for specific
	 * attribute.
	 * @param high
	 * @return List of Integers. 
	 */
	private ArrayList<Integer> FixCombinations(String high)
	{
		ArrayList<Integer> fixedCombinations = new ArrayList<Integer>();
		for(int i=0; i<high.length(); i++)
		{
			fixedCombinations.add(Integer.parseInt(high.charAt(i)+""));
		}
		return fixedCombinations;
	}

	/**
	 * This method is used for two purposes so it is called Magic. 
	 * This method generates elementary rules for specific parameters.
	 * It can check if everything is covered from a group of elementary rules and
	 * it can create new low argument.
	 * @param low
	 * @param change
	 * @param nr
	 * @param combinations
	 * @param method
	 * @return Returns true if complex rules are complete and it returns new low 
	 * bound if a specific does not cover all rules.
	 */
	private String Magic(String low, int change, int nr, ArrayList<Integer> combinations, int method)
	{
		StringBuilder sb = new StringBuilder(low);
		int pos = low.length()-1;
		int tmpPos = pos;
		for(int i=1; i<=nr; i++)
		{	
			switch (method) 
			{
				case 0: 
					if(!IsRule(sb.toString()))
						return "0";
					break;
				case 1:
					if(IsRule(sb.toString()) && i!= 1)
						return sb.toString();
				default:
					break;
			}						
			sb.replace(pos, pos+1, (Integer.parseInt(sb.charAt(pos)+"")+1)+"");
			if((i % change)==0) 
			{	
				for(int j=low.length()-1; j>0; j--)
				{
					if(Integer.parseInt((sb.charAt(j)+"")) >= combinations.get(j))
						tmpPos = j-1;					
					else
						break;
				}
				sb.replace(tmpPos, tmpPos+1, (Integer.parseInt(sb.charAt(tmpPos)+"")+1)+"");
				for(int j=low.length()-1; j>tmpPos; j--)
					sb.replace(j, j+1, (Integer.parseInt(low.charAt(j)+""))+"");	
			}				
		}
		return "1";
	}

	/**
	 * Checks if a specific rule exist in elementary rules.
	 * @param rule
	 * @return True if this rule exists in elementary rules and false
	 * if it does not exists in elementary rules.
	 */
	private boolean IsRule(String rule)
	{
		for(int i=0; i<this.elementaryRules.size(); i++)
		{
			String rule1 = this.elementaryRules.get(i).getRule();
			rule1 = rule1.substring(0, rule1.length()-1);
			if(rule1.equals(rule))
				return true;
		}
		return false;
	}

	/**
	 * This method starts from a low bound to generate rules. 
	 * @param low
	 * @param change
	 * @param nr
	 * @param combinations
	 * @return List of generated rules.
	 */
	private ArrayList<String> CountAndFind(String low, int change, int nr, ArrayList<Integer> combinations)
	{
		ArrayList<String> complexRules = new ArrayList<String>();		
		StringBuilder sb = new StringBuilder(low);
		int pos = low.length()-1;
		int tmpPos = pos;
		for(int i=1; i<=nr; i++){		
			if(IsElementaryRule(sb.toString()))
				complexRules.add(sb.toString());			
			sb.replace(pos, pos+1, (Integer.parseInt(sb.charAt(pos)+"")+1)+"");
			if((i % change)==0) {	
				for(int j=low.length()-1; j>0; j--){
					if(Integer.parseInt((sb.charAt(j)+"")) >= combinations.get(j))
						tmpPos = j-1;					
					else
						break;
				}
				sb.replace(tmpPos, tmpPos+1, (Integer.parseInt(sb.charAt(tmpPos)+"")+1)+"");
				for(int j=low.length()-1; j>tmpPos; j--)
					sb.replace(j, j+1, (Integer.parseInt(low.charAt(j)+""))+"");	
			}				
		}
		return complexRules;
	}

	/**
	 * This method finds a rule in elementary rules and changes 
	 * value of found to true so this rule is covered.
	 * @param rule
	 * @return True if rule is found and covered. False if rule is
	 * not found.
	 */
	private boolean IsElementaryRule(String rule)
	{
		for(int i=0; i<this.elementaryRules.size(); i++)
		{
			String rule1 = this.elementaryRules.get(i).getRule();
			if(rule1.substring(0, rule1.length()-1).equals(rule)){
				this.elementaryRules.get(i).setFound(true);
				return true;
			}
		}
		return false;
	}

	/**
	 * This method translates encoded complex rules and prepares for
	 * printing.
	 * @param complexRules
	 * @param target
	 * @param ruleLength
	 */
	private void PrintComplexRules(ArrayList<String> complexRules, String target, int ruleLength)
	{
		ArrayList<String> tmpComplexRules = new ArrayList<String>();
		String tmpRule = "";
		String complex = "";
		String formatedComplexRule = "";
		for(int i=0; i<ruleLength; i++)
		{
			for(int j=0; j<complexRules.size(); j++)
			{
				tmpRule += complexRules.get(j).charAt(i);
			}
			tmpComplexRules.add(tmpRule);
			tmpRule = "";
		}
		for(int i=0; i<tmpComplexRules.size(); i++)
		{
			tmpRule = tmpComplexRules.get(i);
			tmpRule = Uniques(tmpRule);
			int scale = this.attribute.getAttribute(i).getScaleSize();
			if(tmpRule.length() == 1)
			{
				complex += this.attribute.getAttribute(i).getScaleValue(Integer.parseInt(tmpRule));
			}
			else if(tmpRule.length() == scale)
			{
				complex += "*";
			}		
			else
			{
				String[] minAndMax = findMinAndMax(tmpRule);
				int min = Integer.parseInt(minAndMax[0]);
				int max = Integer.parseInt(minAndMax[1]);
				if(min == 0)
				{
					complex += "<="+this.attribute.getAttribute(i).getScaleValue(max);
				}
				else if(max+1 == scale)
				{
					complex += ">="+this.attribute.getAttribute(i).getScaleValue(min);
				}
				else
				{
					complex += this.attribute.getAttribute(i).getScaleValue(min)+":"+this.attribute.getAttribute(i).getScaleValue(max);
				}
			}
			formatedComplexRule += String.format("%1$-15s", complex);
			complex = "";
		}
		formatedComplexRule += String.format("%1$-15s", "|"+this.attribute.getScaleValue(Integer.parseInt(target))+"|");
		System.out.println(String.format("%1$-4s", complexRulesCounter+")")+formatedComplexRule);
		complexRulesCounter++;
	}
	
	/**
	 * This method finds unique values of a specific attribute of a rule.
	 * @param rule
	 * @return String of unique values.
	 */
	private String Uniques(String rule)
	{
		String temp = "";
	    for (int i = 0; i < rule.length(); i++){
	        char current = rule.charAt(i);
	        if (temp.indexOf(current) < 0){
	            temp += current;
	        }
	    }
	    if(temp.length() == 0)
	    	return rule.charAt(0)+"";
	    else
	    	return temp;
	}
	
	/**
	 * For specific complex rules it finds minimum and maximum
	 * value used to translate into interval for complex rules.
	 * Example: String s = 123 => return 1 as min and 3 as max, if for this
	 * argument 4 is the Scale Size it outputs this rule as >=1
	 * @param rule
	 * @return Min and Max value as array
	 */
	private String[] findMinAndMax(String rule)
	{		
		int value = Integer.parseInt(rule.charAt(0)+"");
		int maxValue = value;
		int minValue = value;		
		for(int i=1;i < rule.length(); i++)
		{ 
			value = Integer.parseInt(rule.charAt(i)+"");
			if(value > maxValue)
				maxValue = value;		      
			if(value < minValue)
		        minValue = value;		      
		} 
		return new String[]{minValue+"",maxValue+""};
	}
}
