////////////////////////////////////////////////////////////////////////////////
//JDEXi3:   Implements evaluation of decision alternatives based on
//          qualitative multi-attribute models produced by DEXi software
//          (http://kt.ijs.si/MarkoBohanec/dexi.html)
//
//          Authors: Marko Bohanec, Dusan Omercevic, Andrej Kogovsek
//          (http://kt.ijs.si/MarkoBohanec/jdexi.html)
//
//JDEXi3 library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.
//
//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//Lesser General Public License for more details.
//
//You should have received a copy of the GNU Lesser General Public
//License along with this library; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
////////////////////////////////////////////////////////////////////////////////

package test.JDEXi;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Random;

import si.JDEXi.NewComplexRules;
import si.JDEXi.Attribute;
import si.JDEXi.ComplexRules;
import si.JDEXi.Model;
import si.JDEXi.Rule;

/**
 * <p>Test JDEXi3 library and show how to program typical operations on DEXi models</p>
 * <p> Parameter: .dxi file name</p>
 * <p>Processing:
 * <ol>
 *   <li>load the model</li>
 *   <li>print lists: basic, aggregate and linked attributes</li>
 *   <li>print attribute structure with descriptions</li>
 *   <li>print attribute structure with scales</li>
 *   <li>print root attribute and its descendants</li>
 *   <li>print root function, if exists: rules and admissible bounds</li>
 *   <li>create a random alternative and evaluate it</li>
 *   <li>print evaluation results, quickly and with attribute structure</li>
 *  </ol>
 *  </p>   
 * 
 * @author Marko Bohanec
 * @version 3.0
 */
public final class TestJDEXi4 {

    /**
     * Main class
     * 
     * @param args
     *            Input parameters
     */
    public static void main(String[] args) {
        try {
        	if (args.length != 1) {
                System.out
                        .println("Usage: java -jar TestJDEXi.jar input_DEXi_model.dxi");
                System.exit(1);
            }

            // load the model

            // old way
            // Model model = new Model(Model.loadFile(args[0]));

            // new way
    
        	File dexiFile = new File(args[0]);
            Model model = new Model(dexiFile);

            // print lists: basic, aggregate and linked attributes
            System.out.println("\nAttributes:");
            System.out.println("  Basic: "
                    + model.attributesToString(model.basic));
            System.out.println("  Aggregate: "
                    + model.attributesToString(model.aggregate));
            System.out.println("  Linked: "
                    + model.attributesToString(model.linked));

            
            
            ArrayList<Attribute> all = model.getAllAttributes();
            
            // calculate maxnl = maximum number of character occupied by attribute name + level indent
            int maxnl = 0;
            for (int i = 0; i < all.size(); i++) {
                int nl = all.get(i).getName().length() + 2
                        * all.get(i).getLevel();
                if (nl > maxnl) {
                    maxnl = nl;
                }
            }
            maxnl = maxnl + 1; // + gap

            // print attribute structure with descriptions
            System.out.println("\nAttributes and descriptions:");

            for (int i = 0; i < all.size(); i++) {
                Attribute att = all.get(i);
                String name = att.getName();
                int len = name.length();
                int lev = 2 * att.getLevel();
                System.out.println(emptyStr(lev) + name
                        + emptyStr(maxnl - len - lev) + att.getDescription());
            }

            // print attribute structure with scales
            System.out.println("\nAttributes and scales:");

            for (int i = 0; i < all.size(); i++) {
                Attribute att = all.get(i);
                String name = att.getName();
                int len = name.length();
                int lev = 2 * att.getLevel();
                String scl = "";
                for (int s = 0; s < att.getScaleCount(); s++) {
                    String val = att.getScaleValue(s);
                    if (scl != "") {
                        scl = scl + ";";
                    }
                    scl = scl + val;
                }
                System.out.println(emptyStr(lev) + name
                        + emptyStr(maxnl - len - lev) + scl);
            }
            
            // print root attribute and its descendants
            Attribute root = all.get(0);
            if (root != null) {
                System.out
                        .println("\nRoot attribute " + root.getName() + " has "
                                + root.getAttributeCount() + " attribute(s):");
                for (int i = 0; i < root.getAttributeCount(); i++) {
                    Attribute a = root.getAttribute(i);
                    System.out.println("  Attribute " + i + "=" + a.getName());
                }

                // print root function, if exists
                if (root.getRuleCount() >= 0) {
                    dumpFunction(root);
                    dumpBounds(root);
                }
                System.out.println();
                System.out.println("Complex Rules for attribute " + root.getName());
                System.out.println();
                
                //Create complex rules for attribute at specific position
                ComplexRules cr = new ComplexRules(root,false);
                cr.createComplexRule();
                //Create new complex rules for specific attribute
                System.out.println();
                System.out.println("New Complex Rules for attribute " + root.getName());
                System.out.println();
                NewComplexRules ncr = new NewComplexRules(root);
                ncr.createNewComplexRules();
                
                // create a random alternative
                System.out.println("\nEvaluating a random alternative");
                Random rnd = new Random();
                model.clearInputValues();
                for (int i = 0; i < model.basic.size(); i++) {
                    model.setInputValue(i,
                            rnd.nextInt(model.basic.get(i).getScaleSize()));
                }

                // evaluate
                model.evaluate(Model.Evaluation.SET, true);

                // print evaluation results
                // ... quickly
                System.out.println("\nInput values:  "
                        + model.getInputValues());
                System.out.println("Output values: "
                        + model.getOutputValues());

                // ... with attribute structure
                System.out.println("\nAggregate attributes and values:");
                for (int i = 0; i < model.aggregate.size(); i++) {
                    Attribute att = model.aggregate.get(i);
                    String name = att.getName();
                    int len = name.length();
                    int lev = 2 * att.getLevel();
                    System.out.println(emptyStr(lev) + name
                            + emptyStr(maxnl - len - lev)
                            + model.getOutputDistrString(i));
                }
            }
        } catch (Exception e) {
            System.out.println("Exception: " + e.getMessage());
        }
    }

    /**
     * Print a.function and demonstrate how to access various attribute, scale
     * and function properties.
     * 
     * @param a
     */
    public static void dumpFunction(Attribute a) {
        System.out.print("\nFunction of " + a.getName() + " depends on:");
        for (int j = 0; j < a.getAttributeCount(); j++) {
            Attribute x = a.getAttribute(j);
            System.out.print(" " + x.getName());
        }
        System.out.println();

        for (int r = 0; r < a.getRuleCount(); r++) {
            Rule rule = a.getRule(r);
            int[] args = a.indexToArgs(r);
            Rule ruleByArgs = a.functionValue(args);
            String msg = (rule == ruleByArgs) ? " " : " !error! ";
            System.out.print(r + ": " + Arrays.toString(args) + " ");
            System.out.print(rule.getLow() + ":" + rule.getHigh() + msg);
            for (int j = 0; j < a.getAttributeCount(); j++) {
                Attribute x = a.getAttribute(j);
                System.out.print(" " + x.getScaleValue(args[j]));
            }
            System.out.println(" -> " + a.getScaleValue(rule.getLow()) + ":"
                    + a.getScaleValue(rule.getHigh()));
        }
    }

    /**
     * Print admissible value bounds of rules and check the consistency of
     * rules. Illustrates the use of ruleBounds(), ruleIsConsistent() and
     * functionIsConsistent().
     * 
     * @param a
     */
    public static void dumpBounds(Attribute a) {
        System.out.println("\nAdmissible bounds of function " + a.getName());

        for (int r = 0; r < a.getRuleCount(); r++) {
            Rule rule = a.getRule(r);
            Rule bounds = a.ruleBounds(r);
            int[] args = a.indexToArgs(r);
            System.out.print(r + ": " + Arrays.toString(args) + " ");
            System.out.print(rule.getLow() + ":" + rule.getHigh());
            System.out.print("  [" + bounds.getLow() + ":" + bounds.getHigh()
                    + "]");
            if (!a.ruleIsConsistent(r, bounds)) {
                System.out.print(" inconsistent");
            }
            System.out.println();
        }
        if (!a.functionIsConsistent()) {
            System.out.println("Warning: function is inconsistent");
        }
    }

    public static String emptyStr(int n) {
        if (n<=0) {return "";}
        else {return String.format("%"+n+"s"," ");}
    }
    
}
